classdef Prensilia < handle
    % DEMO hands class. Provides quick and structured access to devices
    % functions from MATLAB.
    %
    % ----- COMPATIBILITY -----
    % Working with firmware versions from November 2014
    %
    % ----- USAGE EXAMPLES -----
    % IH2 = Prensilia('IH2Azzurra', 'COM1'); % creates a "IH2 Azzurra" object
    %                                        % connected through serial port 1
    %                                        % on a Windows PC
    %
    % EH1 = Prensilia('EH1Milano', '/dev/tty.serial1');
    %                                      % creates a "EH1 Milano" object
    %                                      % connected through serial port
    %                                      % /dev/tty.serial1
    %                                      % on a machine running MacOS
    %
    % ----- Copyright notes -----
    % Copyright (C) 2014,2015 - Scuola Superiore Sant'Anna, Pisa, Italy
    %
    % This program is free software: you can redistribute it and/or modify
    % it under the terms of the GNU General Public License as published by
    % the Free Software Foundation, either version 3 of the License, or
    % (at your option) any later version.
    %
    % This program is distributed in the hope that it will be useful,
    % but WITHOUT ANY WARRANTY; without even the implied warranty of
    % MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    % GNU General Public License for more details.
    %
    % You should have received a copy of the GNU General Public License
    % along with this program.  If not, see <http://www.gnu.org/licenses/>.
    % 
    % ----- TBD -----
    % - add check for input arguments to be numbers only
    % - disable HLHC methods for SAM
    %
    % ----- CHANGELOG -----
    % 15-03-14 (FC) | v.1.0 | command set checked
    % 15-02-14 (FC) | v.0.5 | command set completed (to be checked)
    % 14-06-26 (FC) | v.0.1 | started writing
    
    properties (SetAccess = private) % other classes can't modify these options
        SerialPort = [];
        HWtype = [];
        n_DoA  = [];
        tStart = [];
        Streaming = struct('type',[], 'period',[], 'packet_size',[],...
            'syncing',[]);
        DataLogger = [];
        LoggedData = struct('timeStamp',[], 'position',[],...
            'current',[], 'ExternalSensors',[]);
    end
        
    methods(Access=private, Hidden)
        
   %%%% SUPPORT Functions %%%%
        
        function checkMA(obj,MA)
            % check if the motor address provided to a function is
            % within the allowed range
            switch obj.HWtype
                case 'SAM'
                    if (MA < 0)
                        error('Invalid Motor Address provided');
                    end
                    
                case 'IH2Azzurra'
                    if (MA > 4 || MA < 0)
                        error('Invalid Motor Address provided');
                    end
                    
                case 'EH1Milano'
                    if MA > 5 || MA < 0
                        error('Invalid Motor Address provided');
                    end
            end
        end
        
        function flushSerialPort(obj)
            % flush serial port avoiding the use of external toolboxes
            if (obj.SerialPort.BytesAvailable > 0)
                fread(obj.SerialPort, obj.SerialPort.BytesAvailable);
            end
        end
        
        function P = checkPreshapePosture(obj,P)
            if ((~isvector(P)) || (length(P) ~= obj.n_DoA))
                error(['Incorrect number of positions provided. The '...
                    'Preshape/Posture argument must be a vector of '...
                    num2str(obj.n_DoA), ' elements',])
            end
          
            if sum(P>255) > 0 || sum(P<0) > 0
                warning('PRE:PS_OOR', ['At least one of the provided positions is out of '...
                    'the allowed range (0:255). The closer range limit '...
                    'will be used instead']);
                P = min(P, 255); % works with vectors as well
                P = max(P, 0);
            end
 
            if iscolumn(P), P = P'; end
        
        end
        
        function TC = checkTensionCurrent(obj,TC)
            if ((~isvector(TC)) || (length(TC) ~= obj.n_DoA-1))
                error(['Incorrect number of tensions/currents provided. The '...
                    'Current argument must be a vector of %d '...
                    'elements',(obj.n_DoA-1)])
            end
          
            if sum(TC>1024) > 0 || sum(TC<0) > 0
                warning('PRE:TC_OOR', ['At least one of the provided tensions/currents is out of '...
                    'the allowed range (0:1023). The closer range limit '...
                    'will be used instead']);
                TC = min(TC, 1023); % works with vectors as well (i.e. affects outliers only)
                TC = max(TC, 0);
            end
 
            if iscolumn(TC), TC = TC'; end
            
        end
        
        function PID = checkPIDparams(~,PID)
            if ((~isvector(PID)) || (length(PID) ~= 4))
                error(['Incorrect number of parameters provided. The '...
                    'PIDParams argument must be a vector of 4 '...
                    'elements'])
            end
          
            if sum(PID>255) > 0 || sum(PID<0) > 0
                warning('PRE:PID_OOR', ['At least one of the provided PID parameters is out of '...
                    'the allowed range (0:255). The closer range limit '...
                    'will be used instead']);
                PID = min(PID, 255); % works with vectors as well (i.e. affects outliers only)
                PID = max(PID, 0);
            end
 
            if iscolumn(PID), PID = PID'; end
        end
        
    end %methods(Access=private, hidden)
    
    methods(Access=public)
        
        %%% CONSTRUCTOR
        function obj = Prensilia(varargin)
            % the constructor function builds up the class instance
            % arguments: 1) prototype series
            %            2) serial port number
            %
            % allowed values: 1) - 'SAM'
            %                    - 'IH2Azzurra'
            %                    - 'EH1Milano'
            %                 2) - 'COMn' where n is the serial port number (WINDOWS)
            %                    - '/dev/tty.KeySerial' were KeySerial is the serial port name (MACOS)
            %                    - '/dev/ttyS0' where ttyS0 is the serial port name (LINUX)

            if nargin>2,
                error('Too many input arguments')
            end
            
            % set default values
            DefaultArgValues = {'IH2Azzurra', 'fake'};
            % and eventually replace with provided arguments
            DefaultArgValues(1:size(varargin,2)) = varargin;
            [obj.HWtype, COMport] = deal(DefaultArgValues{1:2});

            
            % Serial communication initialization (depending on hardware)
            switch obj.HWtype
                case 'SAM'
                    BaudRate  = 115200;
                    obj.n_DoA = 1;
           
                case 'IH2Azzurra'
                    BaudRate  = 115200;
                    obj.n_DoA = 5;

                case 'EH1Milano'
                    BaudRate  = 115200;
                    obj.n_DoA = 6;
                    
                otherwise
                    error(['Invalid hardware provided. Supported values are:'...
                          ' ''SAM'', ''IH2Azzurra'', ''EH1Milano'' ']);
            end

            BitsN = 8;
            SBits = 1;
            Parity = 'none';
            TimeOut = 0.2;
            InputBufferSize = 500;
            OutputBufferSize = 500;

            if ~strcmp(COMport, 'fake') % if hardware is attached
                % Open serial port
                try
                    obj.SerialPort = serial(COMport,'BaudRate', BaudRate,...
                        'DataBits', BitsN, 'StopBits', SBits, 'Parity', Parity,...
                        'Timeout', TimeOut, 'InputBufferSize', InputBufferSize,...
                        'OutputBufferSize', OutputBufferSize);
                    fopen(obj.SerialPort);
                catch ME
                    error(ME.identifier,'Invalid serial port supplied!');
                end
            else
                obj.SerialPort = [];
                warning('PRE:Costructor:FakeSerialPort',...
                    'fake serial port provided. No communication will be possible');
            end
            
        end % CONSTRUCTOR
        
        %%% DESTRUCTOR
        function delete(obj)
            
            % Delete serial port (if not 'fake')
            if ~isempty(obj.SerialPort)
                obj.StopDataLog();
                
                % try to close the serial port
                fclose(obj.SerialPort);
                if strcmp(obj.SerialPort.Status, 'closed') &&...
                   strcmp(obj.SerialPort.RecordStatus, 'off')% if successful, do nothing
                else % return an error
                    warning('PRE:Destructor:CloseSerialPort',...
                        ['Serial port was not successfully closed and'...
                        ' it may not be available for subsequent applications.']);
                end
                
                delete(obj.SerialPort);
            end
            
            % Delete object
            clear obj
        end % DESTRUCTOR            

   %%%% HLHC Commands %%%%
                
        function MoveMotor(obj, MA, PWM)
            % Moves one motor at a desired speed and direction
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) PWM: motor speed (-511:511)
            
            obj.checkMA(MA);

            if PWM > 511 || PWM < -511
                warning('PRE:PWM_OOR',...
                    ['The provided PWM is out of the allowed range (-511:511). '...
                    'The closer range limit will be used instead']);
                PWM = min(PWM, 511);
                PWM = max(PWM, -511);
            end
                        
            if PWM >=0
                PWM_bin = dec2bin(PWM,9);
                bitS    = '1';
            else
                PWM_bin = dec2bin(-PWM,9);
                bitS    = '0';
            end
            
            MA_bin = dec2bin(MA,4);
            bit8   = PWM_bin(1);
            Byte1  = bin2dec(['1' bitS MA_bin '0' bit8]);
            Byte2  = bin2dec(PWM_bin(2:end));

            fwrite(obj.SerialPort, [Byte1 Byte2]);
        end

        function SetFingerPosition(obj, MA, Pos)
            % Moves one DoA towards a desired position (position mode)
            % using calibrated position data format
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) POS: desired position (0:255)
            
            obj.checkMA(MA);
            
            if Pos > 255 || Pos < 0
                warning('PRE:POS_OOR', ...
                    ['The provided position is out of the allowed range (0:255).'...
                    'The closer range limit will be used instead']);
                Pos = min(Pos, 255);
                Pos = max(Pos, 0);
            end
            
            fwrite(obj.SerialPort, [68 MA Pos]);
        end
        
        function SetFingerForce(obj, MA, Force)
            % Moves one DoA (all except for thumb abduction) in a closed
            % loop fashion until desired tendon tension force is reached
            % (Tension mode).
            % AVAILABLE ONLY IF TENDON TENSION FORCE SENSORS ARE INSTALLED!
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) Force: desired force (0:1023)
            
            obj.checkMA(MA);
            
            if Force > 1023 || Force < 0
                warning('PRE:FORCE_OOR',...
                    ['The provided force is out of the allowed range (0:1023).'...
                    'The closer range limit will be used instead']);
                Force = min(Force, 255);
                Force = max(Force, 0);
            end
            
            MA_bin    = dec2bin(MA,4);
            Force_bin = dec2bin(Force,10);
            Byte3     = bin2dec(Force_bin(3:end));
            Byte2     = bin2dec([Force_bin(1:2) '00' MA_bin]);
            
            fwrite(obj.SerialPort, [74 Byte2 Byte3]);
        end

        function Pos = GetFingerPosition(obj,MA)
            % Reads the calibrated position of one DoA and returns it
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [69 MA]);
            Pos = fread(obj.SerialPort, 1);
        end
        
        function Force = GetFingerForce(obj, MA)
            % Reads the tendon tension force sensor of one DoA (all except
            % for thumb abduction/adduction)
            % AVAILABLE ONLY IF TENDON TENSION FORCE SENSORS ARE INSTALLED!
            % Arguments: 1) MA: motor address (limits depend on HW)
            
%             if MA == 0
%                 MA = 1;
%                 warning('PRE:Tens_OOR',...
%                     ['No tendon tension sensor for thumb abduction/adduction.',...
%                      ' MA will be set to 1']);
%             end
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, MA);
            read_bytes = fread(obj.SerialPort, 2);
            
            Force = bitand(3,read_bytes(1))*256 + read_bytes(2);
        end

        function Current = GetMotorCurrent(obj, MA)
            % Reads the motor current of one DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [73 MA]);
            read_bytes = fread(obj.SerialPort, 2);
            
            Current = bitand(3,read_bytes(1))*256 + read_bytes(2);
        end
        
        function ExtSensor = GetExternalSensor(obj, ExtSensNo)
            % Reads one of the external sensors available
            % Arguments: 1) ExtSensNo: External sensor address (0:6)
            
            if ExtSensNo > 6 || ExtSensNo < 0
                warning('PRE:ExtSens_OOR',...
                    ['The external sensor address provided is out of '...
                    'the allowed range (0:6).'...
                    'The closer range limit will be used instead']);
                ExtSensNo = min(ExtSensNo, 6);
                ExtSensNo = max(ExtSensNo, 0);
            end
            
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [77 ExtSensNo]);
            read_bytes = fread(obj.SerialPort, 2);
            
            ExtSensor = bitand(3,read_bytes(1))*256 + read_bytes(2);
        end
        
        
   %%%% Status/Calibrate Commands %%%%
        
        function [StatusByte, StatusStruct] = GetFingerStatus(varargin)
            % Check DoA STATUS byte
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) SHOW: set to 1 if you want to show an
            %               interpretation of the Status Byte on the command
            %               window (optional. Default = 0)
            
            switch nargin
                case 1
                    error('Motor Address not provided');                    
                case 2
                    MA = varargin{2};
                    SHOW = 0;                    
                case 3
                    MA = varargin{2};
                    if isempty(varargin{3})
                        SHOW = 0;
                    else
                        SHOW = varargin{3};
                    end
                otherwise
                    error('Too many input arguments');
            end
            
            obj = varargin{1};
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [75 MA]);
            StatusByte = fread(obj.SerialPort, 1);
            StatusByte = dec2bin(StatusByte,8);
            
            switch StatusByte(1:3)
                case '000'
                    ControlMode = 'STOP';
                case '001'
                    ControlMode = 'PWM/Speed';
                case '010'
                    ControlMode = 'Position';
                case '011'
                    ControlMode = 'Tension (tendon)';
                case '100'
                    ControlMode = 'Current';
                case '110'
                    ControlMode = 'Current/Position';
                case '111'
                    ControlMode = 'COM ERROR';
            end
            
            if StatusByte(4) == '1', ControlOK = 1;
            else ControlOK = 0;
            end
            
            if StatusByte(5) == '1', FCA = 1;
            else FCA = 0;
            end
            
            if StatusByte(6) == '1', FCB = 1;
            else FCB = 0;
            end
            
            if StatusByte(7) == '1', Imax = 1;
            else Imax = 0;
            end
            
            if StatusByte(8) == '1', Moving = 1;
            else Moving = 0;
            end
            
            
            StatusStruct.MotorAddress = MA;
            StatusStruct.ControlMode  = ControlMode;
            StatusStruct.ControlOK    = ControlOK;
            StatusStruct.Open         = FCA;
            StatusStruct.Closed       = FCB;
            StatusStruct.Imax         = Imax;
            StatusStruct.Moving       = Moving;
            
            if SHOW ~= 0
                disp('Control OK | Open | Closed | I > Imax | Moving | Control Mode')
                disp([num2str(ControlOK),'          | ', num2str(FCA),'    | ',...
                      num2str(FCB),'      | ', num2str(Imax),'        | ',...
                      num2str(Moving),'      | ', ControlMode]);
            end
        end % GetFingerStatus
        
        function FirstCalibration(obj)
            % FirstCalibration command starts the automatic calibration
            % routine and stores reference positions in the internal memory
            % of the HLHC, fundamental for HLHC position control
            
            fwrite(obj.SerialPort, 66);
        end
        
        function FastCalibration(obj)
            % This must be the first command to send when the hand is
            % switched-ON. It is necessary for calibrated position control,
            % as it retrieves correct references from the internal memory
            
            fwrite(obj.SerialPort, 70);
        end
        
   %%%% WHOLE HAND COMMANDS %%%%
   
        function HLV = HighLevelVersion(obj)
            % Returns the version of the HLHC (18-byte row string)
            
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, 114);
            HLV = char(fread(obj.SerialPort, 18, 'uchar'))'; %#ok<FREAD>
        end
        
        function StopAll(obj)
            % The prototype will stop all movements and ongoing control
            % loops, once this command is issued: each DoA will then go in
            % STOP mode
            
            fwrite(obj.SerialPort, 65);
        end
        
        function SetHandPosture(obj, Posture)
            % All motors of the hand move towards a desired position, hence
            % posturing the hand.
            % Arguments: 1) Posture: array of desired positions. Correct
            %               dimension depends on hardware
            
            PostureOK = obj.checkPreshapePosture(Posture);
            
            fwrite(obj.SerialPort, [72 PostureOK 72]);
        end
        
        function OpenAll(obj)
            % The prototype will reopen flexion/extension DoAs (i.e. not
            % the thumb abduction/adduction)
            
            fwrite(obj.SerialPort, 76);
        end
        
        function EnableStreaming(obj, type)
            % Starts data straming from hardware
            % Arguments: 1) type: streaming type (1: position and current
            %                                     2: current
            %                                     3: position
            %                                     4: external sensors)
            
            if type > 4 || type < 1
                error('Streaming type not supported');
            end
            
            obj.flushSerialPort();
            obj.Streaming.type    = type;
            obj.Streaming.syncing = 0;
            switch type
                case 1
                    obj.Streaming.period = 20;
                    obj.Streaming.packet_size = 20;
                    
                case 2
                    obj.Streaming.period = 15;
                    obj.Streaming.packet_size = 15;
                    
                case 3
                    obj.Streaming.period = 15;
                    obj.Streaming.packet_size = 10;
                    
                case 4
                    obj.Streaming.period = 15;
                    obj.Streaming.packet_size = 19;
            end
            
            fwrite(obj.SerialPort, [67 type]);
        end
        
        function DisableStreaming(obj)
            % Disables data streaming from hardware
            
            obj.Streaming.type = [];
            fwrite(obj.SerialPort, 71);
            obj.flushSerialPort();
        end
        
   %%%% Automatic grasps %%%%
   
        function AutomaticGrasp(obj, GraspType, GraspForce, GraspSteps)
            % Performs an automatic grasp in a stereotypical,
            % pre-programmed manner.
            % Arguments: 1)GraspType: prehensile grasping form (e.g. lateral, 3)
            %            2)GraspForce: force degree (0:255)
            %            3)GraspSteps: maximum time duration of the grasp
            %                          in multiples of 52ms (15:255)
            
            if ischar(GraspType)
                switch GraspType
                    case 'cylindrical'
                        GT = 4;

                    case 'tri-digit'
                        GT = 3;

                    case 'bi-digit'
                        GT = 2;

                    case 'lateral'
                        GT = 1;

                    case 'tr-digit2'
                        GT = 31;

                    case 'bi-digit2'
                        GT = 21;

                    case 'buffet'
                        GT = 11;

                    case 'three'
                        GT = 6;

                    case 'pistol'
                        GT = 7;

                    case 'thumb-up'
                        GT = 8;

                    case 'relax'
                        GT = 0;

                    otherwise
                        error('Wrong grasp type');
                end
            else if sum(GraspType == [4 3 2 1 31 21 11 6 7 8 0]) == 0
                    error('Wrong grasp type');
                else GT = GraspType;
                end
            end
            
            if GraspForce > 255 || GraspForce < 0
                warning('PRE:GF_OOR',...
                    ['The provided grasp force is out of the '...
                    'allowed range (0:255).'...
                    'The closer range limit will be used instead']);
                GraspForce = min(GraspForce, 255);
                GraspForce = max(GraspForce, 0);
            end
            
            if GraspSteps > 255 || GraspSteps < 15
                warning('PRE:GS_OOR',...
                    ['The provided grasp steps is out of the '...
                    'allowed range (15:255).'...
                    'The closer range limit will be used instead']);
                GraspSteps = min(GraspSteps, 255);
                GraspSteps = max(GraspSteps, 15);
            end
            
            fwrite(obj.SerialPort, [111 GT GraspForce GraspSteps]);
        end

   %%%% HLHC Memory Commands %%%%
   
        function MemCylPreShape(obj, Preshape)
            % Stores in the EEPROM preshape positions for the Cylindrical grasp
            % Arguments: 1) Preshape: 5-length vector containing the
            %                         preshape positions (0:255)
  
            PreshapeOK = obj.checkPreshapePosture(Preshape);
          
            fwrite(obj.SerialPort, [88 PreshapeOK 88]);
        end

        function MemTriPreShape(obj, Preshape)
            % Stores in the EEPROM preshape positions for the Tridigital grasp
            % Arguments: 1) Preshape: 5-length vector containing the
            %                         preshape positions (0:255)
            
            PreshapeOK = obj.checkPreshapePosture(Preshape);
          
            fwrite(obj.SerialPort, [90 PreshapeOK 90]);
        end

        function MemBiPreShape(obj, Preshape)
            % Stores in the EEPROM preshape positions for the Bidigital grasp
            % Arguments: 1) Preshape: 5-length vector containing the
            %                         preshape positions (0:255)
            
            PreshapeOK = obj.checkPreshapePosture(Preshape);

            fwrite(obj.SerialPort, [91 PreshapeOK 91]);
        end
        
        function MemLatPreShape(obj, Preshape)
            % Stores in the EEPROM preshape positions for the Lateral grasp
            % Arguments: 1) Preshape: 5-length vector containing the
            %                         preshape positions (0:255)
            
            PreshapeOK = obj.checkPreshapePosture(Preshape);

            fwrite(obj.SerialPort, [89 PreshapeOK 89]);
        end

        function MemLowCurr(obj, Current)
            % Sets the motor current values for the low force grasp, for all DoAs
            % Arguments: 1) Current: 5-length vector containing the
            %                        motor currents (0:255) 
            
            CurrentOK = obj.checkTensionCurrent(Current);
 
            BinC   = dec2bin(CurrentOK,10);
            BytesH = bin2dec(BinC(:,1:2));
            BytesL = bin2dec(BinC(:,3:end));
            
            fwrite(obj.SerialPort, [110 ...
                BytesH(1) BytesL(1) BytesH(2) BytesL(2) ...
                BytesH(3) BytesL(3) BytesH(4) BytesL(4) 110]);
        end
        
        function MemHighCurr(obj, Current)
            % Sets the motor current values for the high force grasp, for all DoAs
            % Arguments: 1) Current: 5-length vector containing the
            %                        motor currents (0:255) 
            
            CurrentOK = obj.checkTensionCurrent(Current);
 
            BinC   = dec2bin(CurrentOK,10);
            BytesH = bin2dec(BinC(:,1:2));
            BytesL = bin2dec(BinC(:,3:end));
            
            fwrite(obj.SerialPort, [108 ...
                BytesH(1) BytesL(1) BytesH(2) BytesL(2) ...
                BytesH(3) BytesL(3) BytesH(4) BytesL(4) 108]);
        end
        
   %%%% LLMC Commands %%%%
        
   %GENERAL COMMANDS
        function [StatusByte, StatusStruct] = STATUS(varargin)
            % Check DoA LLMC STATUS byte
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) SHOW: set to 1 if you want to show an
            %               interpretation of the Status Byte on the command
            %               window

            switch nargin
                case 1
                    error('Motor Address not provided');                    
                case 2
                    MA = varargin{2};
                    SHOW = 0;                    
                case 3
                    MA = varargin{2};
                    if isempty(varargin{3})
                        SHOW = 0;
                    else
                        SHOW = varargin{3};
                    end
                otherwise
                    error('Too many input arguments');
            end
            
            obj = varargin{1};
            obj.checkMA(MA);
            obj.flushSerialPort();

            fwrite(obj.SerialPort, [95 MA 112 MA]);
            StatusByte = fread(obj.SerialPort, 1);
            StatusByte = dec2bin(StatusByte,8);

            switch StatusByte(1:3)
                case '000'
                    ControlMode = 'STOP';
                case '001'
                    ControlMode = 'PWM/Speed';
                case '010'
                    ControlMode = 'Position';
                case '011'
                    ControlMode = 'Tension (tendon)';
                case '100'
                    ControlMode = 'Current';
                case '110'
                    ControlMode = 'Current/Position';
                case '111'
                    ControlMode = 'COM ERROR';
            end

            if StatusByte(4) == '1', ControlOK = 1;
            else ControlOK = 0;
            end

            if StatusByte(5) == '1', FCA = 1;
            else FCA = 0;
            end

            if StatusByte(6) == '1', FCB = 1;
            else FCB = 0;
            end

            if StatusByte(7) == '1', Imax = 1;
            else Imax = 0;
            end

            if StatusByte(8) == '1', Moving = 1;
            else Moving = 0;
            end
            
            
            StatusStruct.MotorAddress = MA;
            StatusStruct.ControlMode  = ControlMode;
            StatusStruct.ControlOK    = ControlOK;
            StatusStruct.Open         = FCA;
            StatusStruct.Closed       = FCB;
            StatusStruct.Imax         = Imax;
            StatusStruct.Moving       = Moving;
            
            if SHOW ~= 0
                disp('Control OK | Open | Closed | I > Imax | Moving | Control Mode')
                disp([num2str(ControlOK),'          | ', num2str(FCA),'    | ',...
                      num2str(FCB),'      | ', num2str(Imax),'        | ',...
                      num2str(Moving),'      | ', ControlMode]);
            end
        end
        
        function STOP(obj, MA)
            % Stops selected DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            fwrite(obj.SerialPort, [95 MA 113 MA]);
        end
        
        function MemPWMmax(obj, MA, PWM)
            % Stores in the EEPROM maximum PWM for selected MA
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) PWM: PWM to be stored in EEPROM (0:511)
            
            obj.checkMA(MA);
            
            if PWM > 511 || PWM < 0
                warning('PRE:MemPWM_OOR',...
                    ['The provided PWM limit is out of the allowed range (0:511). '...
                    'The closer range limit will be used instead']);
                PWM = min(PWM, 511);
                PWM = max(PWM, 0);
            end
            
            BitPWM = dec2bin(PWM,9);
            ByteH  = bin2dec(BitPWM(1));
            ByteL  = bin2dec(BitPWM(2:end));
            
            fwrite(obj.SerialPort, [95 MA 114 ByteH ByteL MA]);
        end
           
        function MemCURRmax(obj, MA, Current)
            % Stores in the EEPROM maximum PWM for selected MA
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) Current: current to be stored in EEPROM (0:1023)
            
            obj.checkMA(MA);
            
            if Current > 1023 || Current < 0
                warning('PRE:Curr_OOR',...
                    ['The provided Current limit is out of the '...
                    'allowed range (0:1023). '...
                    'The closer range limit will be used instead']);
                Current = min(Current, 1023);
                Current = max(Current, 0);
            end
            
            BitPWM = dec2bin(Current,10);
            ByteH  = bin2dec(BitPWM(1:2));
            ByteL  = bin2dec(BitPWM(3:end));
            
            fwrite(obj.SerialPort, [95 MA 115 ByteH ByteL MA]);
        end
        
        function SETPWM(obj, MA, PWM)
            % Moves one motor at a desired speed and direction
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) PWM: motor speed (-511:511)
            
            obj.checkMA(MA);

            if PWM > 511 || PWM < -511
                warning('PRE:PWM_OOR',...
                    ['The provided PWM is out of the allowed range (-511:511). '...
                    'The closer range limit will be used instead']);
                PWM = min(PWM, 511);
                PWM = max(PWM, -511);
            end
                        
            if PWM >=0
                PWM_bin = dec2bin(PWM,9);
                bitS    = '1';
            else
                PWM_bin = dec2bin(-PWM,9);
                bitS    = '0';
            end
            
            bit8   = PWM_bin(1);
            ByteH  = bin2dec([bitS '000000' bit8]);
            ByteL  = bin2dec(PWM_bin(2:end));

            fwrite(obj.SerialPort, [95 MA 116 ByteH ByteL MA]);
        end
        
        function PWMmax = READPWMmax(obj, MA)
            %Reads from EEPROM maximum PWM for selected MA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 118 MA]);
            read_bytes = fread(obj.SerialPort, 2);
            
            PWMmax = bitand(1,read_bytes(1))*256 + read_bytes(2);
        end
           
        function CURRmax = READCURRmax(obj, MA)
            %Reads from EEPROM maximum PWM for selected MA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 119 MA]);
            read_bytes = fread(obj.SerialPort, 2);
            
            CURRmax = bitand(3,read_bytes(1))*256 + read_bytes(2);
        end
        
        function LLV = LowLevelVersion(obj, MA)
            % Returns the version of the selected LLMC (18-byte row string)
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 64 MA]);
            LLV = char(fread(obj.SerialPort, 18, 'uchar'))'; %#ok<FREAD>
        end
        
   %POSITION COMMANDS
   
        function SETP(obj, MA, Position)
            % Moves one motor to a desired position
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) Position: desired position (0:131071)
            
            obj.checkMA(MA);

            if Position > 131071 || Position < 0
                warning('PRE:LLPOS_OOR',...
                    ['The provided position is out of the allowed '...
                    'range (0:131071). '...
                    'The closer range limit will be used instead']);
                Position = min(Position, 131071);
                Position = max(Position, 0);
            end
            
            BitPos = dec2bin(Position, 17);
            bit16  = BitPos(1);
            ByteU  = 33 + bin2dec(bit16)*16;
            ByteH  = bin2dec(BitPos(2:9));
            ByteL  = bin2dec(BitPos(10:end));

            fwrite(obj.SerialPort, [95 MA ByteU ByteH ByteL MA]);
        end
        
        function Position = READP(obj, MA)
            % Reads LLMC position of one DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 34 MA]);
            read_bytes = fread(obj.SerialPort, 3);
            
            Position = bitand(1,read_bytes(1))*65536 + ...
                read_bytes(2)*256 + read_bytes(3);
        end
        
        function ZEROP(obj, MA)
            % Sets the current position as the 0-reference for the selected
            % DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            fwrite(obj.SerialPort, [95 MA 35 MA]);
        end
        
        function PIDP(obj, MA, PIDparams)
            % Stores position PID parameters in the selected EEPROM
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) PIDParams: Kp,Ki,Kd,Error (0:255)
            
            obj.checkMA(MA);
            PID = obj.checkPIDparams(PIDparams);
            
            fwrite(obj.SerialPort, [95 MA 36 PID MA 170]);
        end
        
        function PIDparams = DUMPP(obj, MA)
            % Reads position PID parameters for the selected DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 37 MA]);
            read_bytes = fread(obj.SerialPort, 4);
            
            PIDparams.Kp  = read_bytes(1);
            PIDparams.Ki  = read_bytes(2);
            PIDparams.Kd  = read_bytes(3);
            PIDparams.Err = read_bytes(4);
        end
        
        
   %TENSION COMMANDS (AVAILABLE ONLY IF TENDON TENSION FORCE SENSORS ARE INSTALLED!)
        
        function SETT(obj, MA, Tension)
            % Moves one motor up to a desired tension
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) Tension: desired tension (0:1023)
            
            obj.checkMA(MA);

            if Tension > 1023 || Tension < 0
                warning('PRE:TENS_OOR',...
                    ['The provided tension is out of the allowed '...
                    'range (0:1023). '...
                    'The closer range limit will be used instead']);
                Tension = min(Tension, 1023);
                Tension = max(Tension, 0);
            end
            
            BitT  = dec2bin(Tension, 10);
            ByteH = bin2dec(BitT(1:2));
            ByteL = bin2dec(BitT(3:end));

            fwrite(obj.SerialPort, [95 MA 65 ByteH ByteL MA]);
        end
        
        function Tension = READT(obj, MA)
            % Reads LLMC tension of one DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 66 MA]);
            read_bytes = fread(obj.SerialPort, 2);
            
            Tension = bitand(3,read_bytes(1))*256 + read_bytes(2);
        end
        
        function ZEROT(obj, MA)
            % Sets the current tension as the 0-reference for the selected
            % DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            fwrite(obj.SerialPort, [95 MA 67 MA]);
        end
        
        function PIDT(obj, MA, PIDparams)
            % Stores tension PID parameters in the selected EEPROM
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) PIDParams: Kp,Ki,Kd,Error (0:255)
            
            obj.checkMA(MA);
            PID = obj.checkPIDparams(PIDparams);
            
            fwrite(obj.SerialPort, [95 MA 68 PID MA 170]);
        end
        
        function PIDparams = DUMPT(obj, MA)
            % Reads tension PID parameters for the selected DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 69 MA]);
            read_bytes = fread(obj.SerialPort, 4);
            
            PIDparams.Kp  = read_bytes(1);
            PIDparams.Ki  = read_bytes(2);
            PIDparams.Kd  = read_bytes(3);
            PIDparams.Err = read_bytes(4);
        end
        
        
   %CURRENT COMMANDS
        
        function SETCURR(obj, MA, Current)
            % Moves one motor up to a desired current
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) Current: desired current (0:1023)
            
            obj.checkMA(MA);

            if Current > 1023 || Current < 0
                warning('PRE:CURR_OOR',...
                    ['The provided current is out of the allowed '...
                    'range (0:1023). '...
                    'The closer range limit will be used instead']);
                Current = min(Current, 1023);
                Current = max(Current, 0);
            end
            
            BitC  = dec2bin(Current, 10);
            ByteH = bin2dec(BitC(1:2));
            ByteL = bin2dec(BitC(3:end));

            fwrite(obj.SerialPort, [95 MA 97 ByteH ByteL MA]);
        end
        
        function Current = READCURR(obj, MA)
            % Reads LLMC current of one DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 98 MA]);
            read_bytes = fread(obj.SerialPort, 2);
            
            Current = bitand(3,read_bytes(1))*256 + read_bytes(2);
        end
        
        function ZEROCURR(obj, MA)
            % Sets the current current as the 0-reference for the selected
            % DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            fwrite(obj.SerialPort, [95 MA 99 MA]);
        end
        
        function PIDCURR(obj, MA, PIDparams)
            % Stores current PID parameters in the selected EEPROM
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) PIDParams: Kp,Ki,Kd,Error (0:255)
            
            obj.checkMA(MA);
            PID = obj.checkPIDparams(PIDparams);
            
            fwrite(obj.SerialPort, [95 MA 100 PID MA 170]);
        end
        
        function PIDparams = DUMPCURR(obj, MA)
            % Reads current PID parameters for the selected DoA
            % Arguments: 1) MA: motor address (limits depend on HW)
            
            obj.checkMA(MA);
            obj.flushSerialPort();
            
            fwrite(obj.SerialPort, [95 MA 101 MA]);
            read_bytes = fread(obj.SerialPort, 4);
            
            PIDparams.Kp  = read_bytes(1);
            PIDparams.Ki  = read_bytes(2);
            PIDparams.Kd  = read_bytes(3);
            PIDparams.Err = read_bytes(4);
        end
        
        function SETCURRPOS(obj, MA, Current)
            % Moves one motor up to a desired current. If an obstacle is
            % found, the motor stops
            % Arguments: 1) MA: motor address (limits depend on HW)
            %            2) Current: desired current (0:1023)
            
            obj.checkMA(MA);

            if Current > 1023 || Current < 0
                warning('PRE:CURR_OOR', ['The provided current is out of the allowed '...
                    'range (0:1023). '...
                    'The closer range limit will be used instead']);
                Current = min(Current, 1023);
                Current = max(Current, 0);
            end
            
            BitC  = dec2bin(Current, 10);
            ByteH = bin2dec(BitC(1:2));
            ByteL = bin2dec(BitC(3:end));

            fwrite(obj.SerialPort, [95 MA 102 ByteH ByteL MA]);
        end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
   %%%% Additional functions %%%%
   
        function StartDataLog(obj)
            % creates a timer that continuously reads data streaming and
            % logs it in a structure
            
            obj.tStart = tic;
                
            if isempty(obj.DataLogger) && ~isempty(obj.SerialPort)
                obj.DataLogger = timer('TimerFcn', {@obj.LogData, obj, obj.tStart}, ...
                    'ExecutionMode','fixedRate','Period', obj.Streaming.period/1000);
            end
            start(obj.DataLogger);
        end
        
        function StopDataLog(obj)
            % ends data logging
            
            if ~isempty(obj.DataLogger) && ~isempty(obj.SerialPort)
                stop(obj.DataLogger);
                obj.LoggedData.currentLogType = [];
            end
        end
        
        function LogData(~, ~, ~, obj, start)
            % Logs data streaming from hardware in a structure
            if ~isempty(obj.Streaming.type)
                if obj.Streaming.syncing

                    while(obj.Streaming.syncing && obj.SerialPort.BytesAvailable > 5)

                        read_bytes = fread(obj.SerialPort, 2);
                        if sum(read_bytes' == [85 170]) == 2
                            obj.Streaming.syncing = 0;

                        else
                            while (read_bytes(2) == 85 && obj.SerialPort.BytesAvailable > 5)
                                read_bytes(2) = fread(obj.SerialPort, 1);
                                if read_bytes(2) == 170
                                    obj.Streaming.syncing = 0;
                                end
                            end
                        end

                    end

                else
                    if obj.SerialPort.BytesAvailable >= obj.Streaming.packet_size

                        read_bytes = fread(obj.SerialPort, obj.Streaming.packet_size);
                        read_bytes = read_bytes';

                        if sum(read_bytes(1:3) == [170 85 obj.Streaming.packet_size])==3 &&...
                           sum(read_bytes(end-1:end) == [85 170])==2

                           switch obj.Streaming.type
                               case 1 %current and position
                                   position = read_bytes(6:3:3*(obj.n_DoA-1)+6);
                                   current  = bitand(3, read_bytes(4:3:3*(obj.n_DoA-1)+4))*256 +...
                                              read_bytes(5:3:3*(obj.n_DoA-1)+5);

                                   tEnd = toc(start);
                                   obj.LoggedData(1).timeStamp = cat(1,obj.LoggedData.timeStamp, tEnd);
                                   obj.LoggedData(1).position  = cat(1,obj.LoggedData.position, position);
                                   obj.LoggedData(1).current   = cat(1,obj.LoggedData.current,  current);
                                   obj.LoggedData(1).ExternalSensors = cat(1,obj.LoggedData.ExternalSensors, zeros(1,7));

                               case 2 %current
                                   current  = bitand(3, read_bytes(4:2:2*(obj.n_DoA-1)+4))*256 +...
                                              read_bytes(5:2:2*(obj.n_DoA-1)+5);

                                   tEnd = toc(start);
                                   obj.LoggedData(1).timeStamp = cat(1,obj.LoggedData.timeStamp, tEnd);
                                   obj.LoggedData(1).position  = cat(1,obj.LoggedData.position, zeros(1, obj.n_DoA));
                                   obj.LoggedData(1).current   = cat(1,obj.LoggedData.current,  current);
                                   obj.LoggedData(1).ExternalSensors = cat(1,obj.LoggedData.ExternalSensors, zeros(1,7));

                               case 3 %position
                                   position = read_bytes(4:(obj.n_DoA-1)+4);

                                   tEnd = toc(start);
                                   obj.LoggedData(1).timeStamp = cat(1,obj.LoggedData.timeStamp, tEnd);
                                   obj.LoggedData(1).position  = cat(1,obj.LoggedData.position, position);
                                   obj.LoggedData(1).current   = cat(1,obj.LoggedData.current,  zeros(1, obj.n_DoA));
                                   obj.LoggedData(1).ExternalSensors = cat(1,obj.LoggedData.ExternalSensors, zeros(1,7));

                               case 4 %external sensors
                                   ExtSens = bitand(3, read_bytes(4:2:2*(7-1)+4))*256 +...
                                             read_bytes(5:2:2*(7-1)+5);

                                   tEnd = toc(start);
                                   obj.LoggedData(1).timeStamp = cat(1,obj.LoggedData.timeStamp, tEnd);
                                   obj.LoggedData(1).position  = cat(1,obj.LoggedData.position, zeros(1, obj.n_DoA));
                                   obj.LoggedData(1).current   = cat(1,obj.LoggedData.current,  zeros(1, obj.n_DoA));
                                   obj.LoggedData(1).ExternalSensors = cat(1,obj.LoggedData.ExternalSensors, ExtSens);

                           end

                        else
                            obj.Streaming.syncing = 1;
                        end
                    end
                end
            end
            
        end %LogData

        
    end % methods(Access=public)

end